package FluxY_Lib::MS::ParseQL;

################################################################################
# This module provides support for parsing Mass Spectrometry data generated by
# QuanLynx, in its default report format.
#
################################################################################

use strict;
use warnings;

use FluxY_Lib::Data::Data_Check;    # User Lib - Handles basic data checking.
use FluxY_Lib::MS::ErrorsAndHelp;   # User Lib - Prints GCMS related messages.
use FluxY_Lib::IO::IO_Utils;        # User Lib - Provides I/O routines.

use base 'Exporter';

our @EXPORT = qw(ParseQuanLynxInput);

################################################################################
# Subroutine: ParseQuanLynxInput
#
# Extracts data from a QuanLynx data file. Data is organized by compound.
#
################################################################################
sub ParseQuanLynxInput{
   my ($dataFile, $msType) = @_;

   # Read in data file and remove blank lines.
   my @lines = &ReadFileSkipB($dataFile);
   
   if ( $lines[0] =~ "Compound" ) {
      return ( &ParseCompoundSummary($dataFile, $msType, @lines) );
   } elsif ($lines[0] =~ "Sample") {
      return ( &ParseSampleSummary($dataFile, $msType, @lines) );
   } else {
      &Error(200, $dataFile, "Quanlynx");
   }
}


################################################################################
# Subroutine: ParseCompoundSummary
#
# Extracts data from a QuanLynx Compound Summary data file.
# - Data is organized by Compound.
#
################################################################################
sub ParseCompoundSummary{
   my ($dataFile, $msType, @lines) = @_;
   
   my @cNames    = ();   # The array of compund names.
   my @cMasses   = ();   # The array of compund masses.
   my @sNames    = ();   # The array of sample names.
   my @dataResps = ();   # The array of MS measurements (identified as response values).
   my @dataAreas = ();   # Second MS quantification method (identified as area values).
   my @dataRT    = ();   # The array of retention times (unique for each measurment).
   my @dataFlags = ();   # Detection flags for each data set (QL specific flags).

   # After the title row, each dataset begins with a heading row starting with:
   # 'Compound' and ending with the compound/mass name. The sample data begins on 
   # the second line.
   # This requires the compound name and the mass number to be separated by a '-'.
   my @cIndex = ();
   for (my $i=1; $i < @lines; $i++) {                # Skip the title line
      if ($lines[$i] =~ "Compound") {             # Extract compoud name, mass & index.

         my ($name, $mass) = split (/-/, &GetLastString($lines[$i]));

         push (@cNames,  lc($name));        # Names in lower case for comparison.
         push (@cMasses, $mass);
         push (@cIndex,  ($i+2));           # Offset to first data value in set.
      }
   }

   # Check that at least one dataset is found.
   if (@cIndex == 0) { &Error(201, $dataFile); }

   # The number of samples can be found in the number of lines between any two 
   # data sets (less two for the header rows) if there are two or more, or
   # from the starting index of the last data set to the end of the file - which 
   # only requires a minimum of one data set to be present.
   my $numSamples = @lines - $cIndex[$#cIndex];

   # Checks that all data sets contain the same number of samples.
   for (my $i=1; $i < @cIndex; $i++) {
      if ( $numSamples != ($cIndex[$i] - $cIndex[$i-1] - 2) ) { &Error(202, $dataFile, $i-1); }
   }

   # Extract the sample names from the first data set.
   my @data  = ();
   for (my $i=0; $i<$numSamples; $i++) {
      @data  = split(/\t/, $lines[$cIndex[0]+$i]);
      my ($str) = split(/\./, $data[2]);
      push(@sNames, $str);
   }
   
   # For each sample extract the retention time, area and response values, and
   # detection flags for all monitorred data.
   for (my $i=0; $i<$numSamples; $i++) {
      my @RTs   = ();
      my @Areas = ();
      my @Resps = ();
      my @Flags = ();
      for (my $j=0; $j<@cNames; $j++) {
         @data = split(/\t/, $lines[$cIndex[$j]+$i]);
         
         push(@RTs,   $data[5]);
         push(@Areas, $data[6]);
         push(@Resps, $data[8]);
         push(@Flags, $data[9]);
      }
      push (@dataRT,    [ @RTs ]);
      push (@dataAreas, [ @Areas ]);
      push (@dataResps, [ @Resps ]);
      push (@dataFlags, [ @Flags ]);
   }
   
   return(\@cNames, \@cMasses, \@sNames, \@dataResps, \@dataAreas, \@dataRT, \@dataFlags);
}


################################################################################
# Subroutine: ParseSampleSummary
#
# Extracts data from a QuanLynx Sample Summary data file.
# - Data is organized by sample.
#
################################################################################
sub ParseSampleSummary{
   my ($dataFile, $msType, @lines) = @_;
   
   my @cNames    = ();   # The array of compund names.
   my @cMasses   = ();   # The array of compund masses.
   my @sNames    = ();   # The array of sample names.
   my @dataResps = ();   # The array of MS measurements (identified as response values).
   my @dataAreas = ();   # Second MS quantification method (identified as area values).
   my @dataRT    = ();   # The array of retention times (unique for each measurment).
   my @dataFlags = ();   # Detection flags for each data set (QL specific flags).

   # After the title rows, each dataset begins with a heading row starting with:
   # 'Sample ' and containing the data set name. The sample data begins on 
   # the second line.
   # This requires the compound name and the mass number to be separated by a '-'.
   my $numSamples = 0;
   my @sIndex     = ();
   for (my $i=1; $i < @lines; $i++) {               # Skip the title line
      if ($lines[$i] =~ "Sample ") {             # Extract the sample name.
         $numSamples++;
         
         my @list = split (/\s/, $lines[$i]);
         push (@sNames, $list[2]);
         push (@sIndex, ($i+2));          # Offset to first data value in set.
      }
   }

   # Check that at least one dataset is found.
   if (@sIndex == 0) { &Error(201, $dataFile); }

   # The number of compounds can be found in the number of lines between any two
   # data sets (less two for the header rows) if there are two or more, or
   # from the starting index of the last data set to the end of the file - which
   # only requires a minimum of one data set to be present.
   my $numCompounds = @lines - $sIndex[$#sIndex];

   # Checks that all data sets contain the same number of samples.
   for (my $i=1; $i < @sIndex; $i++) {
      if ( $numCompounds != ($sIndex[$i] - $sIndex[$i-1] - 2) ) { &Error(202, $dataFile, $i-1); }
   }
   
   # Extract the compound name and masses from the first data set.
   my @data   = ();
   for (my $i=0; $i<$numCompounds; $i++) {
      @data  = split(/\t/, $lines[$sIndex[0]+$i]);
      my ($name, $mass) = split(/-/, $data[2]);
      
      push (@cNames,  lc($name));   # Names in lower case for comparison.
      push (@cMasses, $mass);
   }

   # For each sample extract the retention time, area and response values, and
   # detection flags for all monitorred data.
   for (my $i=0; $i<$numSamples; $i++) {
      my @RTs   = ();
      my @Areas = ();
      my @Resps = ();
      my @Flags = ();
      for (my $j=0; $j<@cNames; $j++) {
         @data = split(/\t/, $lines[$sIndex[$i]+$j]);
         
         push(@RTs,   $data[4]);
         push(@Areas, $data[5]);
         push(@Resps, $data[7]);
         push(@Flags, $data[8]);
      }
      push (@dataRT,    [ @RTs ]);
      push (@dataAreas, [ @Areas ]);
      push (@dataResps, [ @Resps ]);
      push (@dataFlags, [ @Flags ]);
   }
   
   return(\@cNames, \@cMasses, \@sNames, \@dataResps, \@dataAreas, \@dataRT, \@dataFlags);
}


1;
